import random

import torch
from torch import distributions, nn
from torch.autograd import Variable
from torch.distributions import Normal
from torch.nn import Parameter
from torch.nn import functional as F


class Encoder(nn.Module):
    def __init__(
        self,
        note_embedding_dim,
        rnn_hidden_size,
        num_layers,
        num_notes,
        dropout,
        bidirectional,
        z_dim,
        rnn_class,
    ):
        super(Encoder, self).__init__()
        self.bidirectional = bidirectional
        self.num_directions = 2 if bidirectional else 1
        self.note_embedding_dim = note_embedding_dim
        self.num_layers = num_layers
        self.rnn_hidden_size = rnn_hidden_size
        self.z_dim = z_dim
        self.dropout = dropout
        self.rnn_class = rnn_class
        self.lstm = self.rnn_class(
            input_size=note_embedding_dim,
            hidden_size=rnn_hidden_size,
            num_layers=num_layers,
            dropout=self.dropout,
            bidirectional=self.bidirectional,
            batch_first=True,
        )
        self.num_notes = num_notes
        self.note_embedding_layer = nn.Embedding(
            self.num_notes, self.note_embedding_dim
        )

        self.linear_mean = nn.Sequential(
            nn.Linear(
                self.rnn_hidden_size * self.num_directions * self.num_layers,
                self.rnn_hidden_size * self.num_directions,
            ),
            nn.SELU(),
            nn.Linear(self.rnn_hidden_size * self.num_directions, z_dim),
        )

        self.linear_log_std = nn.Sequential(
            nn.Linear(
                self.rnn_hidden_size * self.num_directions * self.num_layers,
                self.rnn_hidden_size * self.num_directions,
            ),
            nn.SELU(),
            nn.Linear(self.rnn_hidden_size * self.num_directions, z_dim),
        )

        self.xavier_initialization()

    def __repr__(self):
        """
        String Representation of class
        :return: string, class representation
        """
        return (
            f"Encoder("
            f"{self.note_embedding_dim},"
            f"{self.rnn_class},"
            f"{self.num_layers},"
            f"{self.rnn_hidden_size},"
            f"{self.dropout},"
            f"{self.bidirectional},"
            f"{self.z_dim},"
            f")"
        )

    def xavier_initialization(self):
        """
        Initializes the network params
        :return:
        """
        for name, param in self.named_parameters():
            if "weight" in name:
                nn.init.xavier_normal_(param)

    def hidden_init(self, batch_size):
        """
        Initializes the hidden state of the encoder GRU
        :param batch_size: int
        :return: torch tensor,
        (
            self.num_encoder_layers x self.num_directions,
            batch_size,
            self.encoder_hidden_size
        )
        """
        hidden = torch.zeros(
            (self.num_layers * self.num_directions, batch_size, self.rnn_hidden_size)
        ).to(torch.cuda.current_device())
        return hidden

    def embed_forward(self, score_tensor):
        """
        Performs the forward pass of the embedding layer
        :param score_tensor: torch tensor,
                (batch_size, measure_seq_len)
        :return: torch tensor,
                (batch_size, measure_seq_len, embedding_size)
        """
        x = self.note_embedding_layer(score_tensor)
        return x

    def forward(self, score_tensor):
        """
        Performs the forward pass of the model, overrides torch method
        :param score_tensor: torch Variable
                (batch_size, measure_seq_len)
        :return: torch distribution
        """
        for name, param in self.named_parameters():
            if "weight" in name:
                nan_check = torch.isnan(param.data)
                if nan_check.nonzero().size(0) > 0:
                    print("Encoder has become nan")
                    raise ValueError

        batch_size, measure_seq_len = score_tensor.size()

        # embed score
        embedded_seq = self.embed_forward(score_tensor=score_tensor)

        # pass through RNN
        hidden = self.hidden_init(batch_size)
        _, hidden = self.lstm(embedded_seq, hidden)
        hidden = hidden.transpose(0, 1).contiguous()
        hidden = hidden.view(batch_size, -1)

        # compute distribution parameters
        z_mean = self.linear_mean(hidden)
        z_log_std = self.linear_log_std(hidden)

        # print(z_mean, z_log_std)
        z_distribution = Normal(loc=z_mean, scale=torch.exp(z_log_std))
        return z_distribution


class Decoder(nn.Module):
    def __init__(
        self,
        note_embedding_dim,
        num_notes,
        z_dim,
    ):
        super(Decoder, self).__init__()
        self.name = "DecoderABC"
        self.num_notes = num_notes
        self.note_embedding_dim = note_embedding_dim
        self.z_dim = z_dim
        self.note_embedding_layer = nn.Embedding(
            self.num_notes, self.note_embedding_dim
        )

    def forward(self, z, score_tensor, train=None):
        """
        :param z: torch_tensor, latent variable
        :param score_tensor: torch_tensor, original measure score tensor
        :param train: bool
        :return:
        """
        return None, None

    def check_index(self, indices):
        """
        :param indices: int,
        :return: bool
        """
        indices = indices.cpu()
        if min(indices) >= 0 and max(indices) < self.num_notes:
            return True
        else:
            print("Invalid Values of Indices: ", min(indices), max(indices))
            raise ValueError

    def xavier_initialization(self):
        """
        Initializes the network params
        :return:
        """
        for name, param in self.named_parameters():
            if "weight" in name:
                nn.init.xavier_normal_(param)


class SRDecoder(Decoder):
    def __init__(
        self,
        note_embedding_dim,
        num_notes,
        z_dim,
        num_layers,
        rnn_hidden_size,
        dropout,
        rnn_class,
    ):
        super(SRDecoder, self).__init__(note_embedding_dim, num_notes, z_dim)
        self.name = "SRDecoder"
        self.num_layers = num_layers
        self.rnn_hidden_size = rnn_hidden_size
        self.dropout = dropout

        # define the individual layers
        self.z_to_rnn_input = nn.Sequential(
            nn.Linear(self.z_dim, self.rnn_hidden_size),
            nn.SELU(),
            nn.Linear(self.rnn_hidden_size, self.note_embedding_dim),
        )

        self.x_0 = Parameter(data=torch.zeros(note_embedding_dim))
        self.rnn_class = rnn_class
        self.rnn_dec = self.rnn_class(
            input_size=2 * self.note_embedding_dim,  # input to bear RNN will be zeros
            hidden_size=self.rnn_hidden_size,
            num_layers=self.num_layers,
            dropout=self.dropout,
            batch_first=True,
        )

        self.rnn_out_to_note_emb = nn.Sequential(
            nn.Linear(self.rnn_hidden_size, self.num_notes), nn.ReLU()
        )

        self.use_teacher_forcing = True
        self.teacher_forcing_prob = 0.5
        self.sampling = "argmax"
        self.xavier_initialization()

    def __repr__(self):
        """
        String Representation of class
        :return: string, class representation
        """
        return (
            f"{self.name}"
            f"{self.note_embedding_dim},"
            f"{self.rnn_class},"
            f"{self.num_layers},"
            f"{self.rnn_hidden_size},"
            f"{self.dropout},"
            f")"
        )

    def hidden_init(self, batch_size):
        """
        :param batch_size: int,
        :return: torch tensor,
                (self.num_layers, batch_size, self.rnn_hidden_size)
        """
        h = torch.zeros(self.num_layers, batch_size, self.rnn_hidden_size)
        return h

    def forward(self, z, score_tensor, train):
        """
        :param z: torch tensor,
                (batch_size, self.z_dim)
        :param score_tensor: torch tensor
                (batch_size, measure_seq_len)
        :return: weights: torch tensor,
                (batch_size, measure_seq_len, self.num_notes)
                samples: torch tensor,
                (batch_size, measure_seq_len)
        """
        if self.use_teacher_forcing and train:
            teacher_forced = random.random() < self.teacher_forcing_prob
        else:
            teacher_forced = False
        if not train:
            sampling = "argmax"
        else:
            sampling = self.sampling

        batch_size_z, z_dim = z.size()
        assert z_dim == self.z_dim
        batch_size, measure_seq_len = score_tensor.size()
        assert batch_size == batch_size_z

        # compute output of rnn_dec
        weights, samples = self.forward_rnn(z, score_tensor, teacher_forced, sampling)
        return weights, samples

    def forward_rnn(self, z, score_tensor, teacher_forced, sampling):
        """
        :param z: torch tensor,
                (batch_size, self.z_dim):
        :param score_tensor: torch tensor,
                (batch_size, measure_seq_len)
        :param teacher_forced: bool,
        :param sampling: string
        :return:
        """
        samples = []
        weights = []
        batch_size, measure_seq_len = score_tensor.size()
        hidden = self.hidden_init(batch_size)
        rnn_input = self.x_0.unsqueeze(0).expand(batch_size, self.note_embedding_dim)
        rnn_input = rnn_input.unsqueeze(1)
        rnn_input_emb = self.z_to_rnn_input(z)
        rnn_input_emb = rnn_input_emb.unsqueeze(1)
        for i in range(measure_seq_len):
            rnn_input = torch.cat((rnn_input, rnn_input_emb), 2)
            rnn_out, hidden = self.rnn_dec(rnn_input, hidden)
            probs = self.rnn_out_to_note_emb(rnn_out[:, 0, :])
            # sample and embed next rnn_input
            if self.use_teacher_forcing and teacher_forced:
                indices = score_tensor.detach()[:, i]
                indices = indices.unsqueeze(1)
                assert self.check_index(indices)
            else:
                if sampling == "multinomial":
                    softmax = F.softmax(probs.detach(), dim=1)
                    indices = torch.multinomial(softmax, 1)
                    assert self.check_index(indices)
                elif sampling == "argmax":
                    _, indices = probs.detach().topk(k=1, dim=1)
                    try:
                        self.check_index(indices)
                    except ValueError:
                        print(probs)
                        raise ValueError
                else:
                    raise NotImplementedError
            rnn_input = self.note_embedding_layer(indices)
            samples.append(indices[:, :, None])
            # save all
            probs = probs.view(batch_size, self.num_notes)
            weights.append(probs[:, None, :])
        weights = torch.cat(weights, 1)
        samples = torch.cat(samples, 2)
        return weights, samples


class SRDecoderNoInput(SRDecoder):
    def __init__(
        self,
        note_embedding_dim,
        num_notes,
        z_dim,
        num_layers,
        rnn_hidden_size,
        dropout,
        rnn_class,
    ):
        super(SRDecoderNoInput, self).__init__(
            note_embedding_dim,
            num_notes,
            z_dim,
            num_layers,
            rnn_hidden_size,
            dropout,
            rnn_class,
        )
        self.name = "SRDecoderNoInput"
        # define the individual layers
        self.z_to_rnn_input = nn.Sequential(
            nn.Linear(self.z_dim, self.rnn_hidden_size),
        )

        self.x_0 = Parameter(data=torch.zeros(note_embedding_dim))
        self.rnn_class = rnn_class
        self.rnn_dec = self.rnn_class(
            input_size=self.rnn_hidden_size,  # input to bear RNN will be zeros
            hidden_size=self.rnn_hidden_size,
            num_layers=self.num_layers,
            dropout=self.dropout,
            batch_first=True,
        )

        self.rnn_out_to_note_emb = nn.Sequential(
            nn.Linear(self.rnn_hidden_size, self.num_notes), nn.ReLU()
        )

        self.xavier_initialization()

    def __repr__(self):
        """
        String Representation of class
        :return: string, class representation
        """
        return (
            f"{self.name}"
            f"{self.note_embedding_dim},"
            f"{self.rnn_class},"
            f"{self.num_layers},"
            f"{self.rnn_hidden_size},"
            f"{self.dropout},"
            f")"
        )

    def hidden_init(self, batch_size):
        """
        :param batch_size: int,
        :return: torch tensor,
                (self.num_layers, batch_size, self.rnn_hidden_size)
        """
        h = torch.zeros(self.num_layers, batch_size, self.rnn_hidden_size)
        return h

    def forward_rnn(self, z, score_tensor, teacher_forced, sampling):
        """
        :param z: torch tensor,
                (batch_size, self.z_dim):
        :param score_tensor: torch tensor,
                (batch_size, measure_seq_len)
        :param teacher_forced: bool,
        :param sampling: string
        :return:
        """
        batch_size, measure_seq_len = score_tensor.size()
        hidden = self.hidden_init(batch_size=batch_size)
        rnn_input = self.z_to_rnn_input(z)
        rnn_input = rnn_input.unsqueeze(1).expand(-1, measure_seq_len, -1)
        rnn_out, hidden = self.rnn_dec(rnn_input, hidden)
        rnn_out = rnn_out.contiguous().view(batch_size * measure_seq_len, -1)
        weights = self.rnn_out_to_note_emb(rnn_out)
        weights = weights.contiguous().view(batch_size, measure_seq_len, -1)
        if sampling == "multinomial":
            softmax = F.softmax(weights.detach, dim=2)
            samples = torch.multinomial(softmax, 2)
        elif sampling == "argmax":
            _, samples = weights.detach().topk(k=1, dim=2)
        return weights, samples


class HierarchicalDecoder(Decoder):
    def __init__(
        self,
        note_embedding_dim,
        num_notes,
        z_dim,
        num_layers,
        rnn_hidden_size,
        dropout,
        rnn_class,
    ):
        super(HierarchicalDecoder, self).__init__(note_embedding_dim, num_notes, z_dim)
        self.name = "HierarchicalDecoder"
        self.rnn_class = rnn_class
        self.num_layers = num_layers
        self.rnn_hidden_size = rnn_hidden_size
        self.dropout = dropout

        # define the individual layers
        self.z_to_beat_rnn_input = nn.Sequential(
            nn.Linear(self.z_dim, self.rnn_hidden_size * self.num_layers), nn.SELU()
        )

        self.beat_rnn_input_dim = 1
        self.b_0 = Parameter(data=torch.zeros(self.beat_rnn_input_dim))
        self.rnn_beat = self.rnn_class(
            input_size=self.beat_rnn_input_dim,
            hidden_size=self.rnn_hidden_size,
            num_layers=self.num_layers,
            dropout=self.dropout,
            batch_first=True,
        )

        self.beat_emb_to_tick_rnn_hidden = nn.Sequential(
            nn.Linear(self.rnn_hidden_size, self.rnn_hidden_size * self.num_layers),
            nn.SELU(),
        )

        self.beat_emb_to_tick_rnn_input = nn.Sequential(
            nn.Linear(self.rnn_hidden_size, self.rnn_hidden_size), nn.SELU()
        )

        self.x_0 = Parameter(data=torch.zeros(note_embedding_dim))
        self.rnn_tick = self.rnn_class(
            input_size=self.note_embedding_dim
            + self.rnn_hidden_size,  # input to bear RNN will be zeros
            hidden_size=self.rnn_hidden_size,
            num_layers=self.num_layers,
            dropout=self.dropout,
            batch_first=True,
        )

        self.tick_emb_to_note_emb = nn.Sequential(
            nn.Linear(self.rnn_hidden_size, self.num_notes), nn.ReLU()
        )

        self.use_teacher_forcing = True
        self.teacher_forcing_prob = 0.5
        self.sampling = "argmax"
        self.xavier_initialization()

    def __repr__(self):
        """
        String Representation of class
        :return: string, class representation
        """
        return (
            f"{self.name}"
            f"{self.note_embedding_dim},"
            f"{self.rnn_class},"
            f"{self.num_layers},"
            f"{self.rnn_hidden_size},"
            f"{self.dropout},"
            f")"
        )

    def hidden_init(self, inp, rnn_type):
        """
        Initializes the hidden state based on the RNN type
        :param inp: torch tensor,
                (batch_size, num_feats)
        :param rnn_type: str, 'beat' for beat rnn, 'tick' for tick rnn
        :return: torch tensor,
                (self.num_layers, batch_size, self.rnn_hidden_size)
        """
        batch_size = inp.size(0)
        if rnn_type == "beat":
            h_0 = self.z_to_beat_rnn_input(inp)
        elif rnn_type == "tick":
            h_0 = self.beat_emb_to_tick_rnn_hidden(inp)
        else:
            raise ValueError
        h = h_0.view(batch_size, self.num_layers, -1)
        h = h.transpose(0, 1).contiguous()
        return h

    def forward(self, z, score_tensor, train):
        """
        Performs the forward pass of the model, overrides torch method
        :param z: torch tensor,
                (batch_size, self.z_dim)
        :param score_tensor: torch tensor
                (batch_size, measure_seq_len)
        :return: weights: torch tensor,
                (batch_size, measure_seq_len, self.num_notes)
                samples: torch tensor,
                (batch_size, measure_seq_len)
        """
        for name, param in self.named_parameters():
            if "weight" in name:
                nan_check = torch.isnan(param.data)
                if nan_check.nonzero().size(0) > 0:
                    print("Decoder has become nan")
                    raise ValueError

        if self.use_teacher_forcing and train:
            teacher_forced = random.random() < self.teacher_forcing_prob
        else:
            teacher_forced = False
        if not train:
            sampling = "argmax"
        else:
            sampling = self.sampling

        batch_size_z, z_dim = z.size()
        assert z_dim == self.z_dim
        batch_size, measure_seq_len = score_tensor.size()
        assert batch_size == batch_size_z

        # compute output of beat rnn
        beat_seq_len = 4
        beat_rnn_out = self.forward_beat_rnn(z, beat_seq_len)

        # compute output of tick rnn
        tick_seq_len = 6
        weights, samples = self.forward_tick_rnn(
            score_tensor, beat_rnn_out, tick_seq_len, teacher_forced, sampling
        )

        return weights, samples

    def forward_beat_rnn(self, z, seq_len):
        """
        Computes the forward pass of the Beat RNN
        :param z: torch tensor,
                (batch_size, self.z_dim)
        :param seq_len: int, sequence length for beat RNN unrolling
        :return:
        """
        batch_size = z.size(0)
        hidden = self.hidden_init(z, rnn_type="beat")
        beat_rnn_input = self.b_0.unsqueeze(0).expand(
            batch_size, seq_len, self.beat_rnn_input_dim
        )
        beat_rnn_out, _ = self.rnn_beat(beat_rnn_input, hidden)
        return beat_rnn_out

    def forward_tick_rnn(
        self, score_tensor, beat_rnn_out, tick_seq_len, teacher_forced, sampling
    ):
        """
        Computes the forward pass of the Tick RNN
        :param score_tensor: torch tensor,
                (batch_size, measure_seq_len)
        :param beat_rnn_out: torch tensor,
                (batch_size, beat_seq_len, self.rnn_hidden_size)
        :param tick_seq_len: int, sequence length for tick RNN unrolling
        :param teacher_forced: bool, whether to use teacher forcing or not
        :param sampling: string, which sampling method to use
        :return:
        """
        samples = []
        weights = []
        batch_size, beat_seq_len, _ = beat_rnn_out.size()
        tick_rnn_input = self.x_0.unsqueeze(0).expand(
            batch_size, self.note_embedding_dim
        )
        tick_rnn_input = tick_rnn_input.unsqueeze(1)
        for i in range(beat_seq_len):
            hidden = self.hidden_init(beat_rnn_out[:, i, :], rnn_type="tick")
            beat_emb_input = self.beat_emb_to_tick_rnn_input(
                beat_rnn_out[:, i, :]
            ).unsqueeze(1)
            for j in range(tick_seq_len):
                tick_rnn_input = torch.cat((tick_rnn_input, beat_emb_input), 2)
                tick_rnn_out, hidden = self.rnn_tick(tick_rnn_input, hidden)
                probs = self.tick_emb_to_note_emb(tick_rnn_out[:, 0, :])
                # sample and embed next rnn_input
                if self.use_teacher_forcing and teacher_forced:
                    indices = score_tensor.detach()[:, i * tick_seq_len + j]
                    indices = indices.unsqueeze(1)
                    assert self.check_index(indices)
                else:
                    if sampling == "multinomial":
                        softmax = F.softmax(probs.detach(), dim=1)
                        indices = torch.multinomial(softmax, 1)
                        assert self.check_index(indices)
                    elif sampling == "argmax":
                        _, indices = probs.detach().topk(k=1, dim=1)
                        try:
                            self.check_index(indices)
                        except ValueError:
                            print(probs)
                            raise ValueError
                    else:
                        raise NotImplementedError
                tick_rnn_input = self.note_embedding_layer(indices)
                samples.append(indices[:, :, None])
                # save all
                probs = probs.view(batch_size, self.num_notes)
                weights.append(probs[:, None, :])
        weights = torch.cat(weights, 1)
        samples = torch.cat(samples, 2)
        return weights, samples


class MeasureVAE(nn.Module):
    def __init__(
        self,
        note_embedding_dim=10,
        metadata_embedding_dim=2,
        num_encoder_layers=2,
        encoder_hidden_size=512,
        encoder_dropout_prob=0.5,
        latent_space_dim=256,
        num_decoder_layers=2,
        decoder_hidden_size=512,
        decoder_dropout_prob=0.5,
        has_metadata=False,
    ):
        """
        Initializes the Measure VAE class object
        :param dataset: FolkMeasureDataset object
        :param note_embedding_dim: int,
        :param metadata_embedding_dim: int,
        :param num_encoder_layers: int,
        :param encoder_hidden_size: int,
        :param encoder_dropout_prob: float, from 0. to 1.
        :param latent_space_dim: int,
        :param num_decoder_layers: int,
        :param decoder_hidden_size: int,
        :param decoder_dropout_prob: float, from 0. to 1.
        :param has_metadata: bool
        """
        super(MeasureVAE, self).__init__()
        # define constants
        self.num_beats_per_measure = 4
        self.num_ticks_per_measure = 24
        self.num_ticks_per_beat = int(
            self.num_ticks_per_measure / self.num_beats_per_measure
        )

        # initialize members
        # self.dataset = dataset.__repr__()
        self.note_embedding_dim = note_embedding_dim
        self.metadata_embedding_dim = metadata_embedding_dim
        self.num_encoder_layers = num_encoder_layers
        self.encoder_hidden_size = encoder_hidden_size
        self.encoder_dropout_prob = encoder_dropout_prob
        self.latent_space_dim = latent_space_dim
        self.num_decoder_layers = num_decoder_layers
        self.decoder_hidden_size = decoder_hidden_size
        self.decoder_dropout_prob = decoder_dropout_prob
        self.has_metadata = has_metadata
        self.num_notes = 130  # len(dataset.note2index_dicts[0])
        print("NUMBER OF NOTES: ", self.num_notes)
        # Encoder
        self.encoder = Encoder(
            note_embedding_dim=self.note_embedding_dim,
            rnn_hidden_size=self.encoder_hidden_size,
            num_layers=self.num_encoder_layers,
            num_notes=self.num_notes,
            dropout=self.encoder_dropout_prob,
            bidirectional=True,
            z_dim=self.latent_space_dim,
            rnn_class=torch.nn.GRU,
        ).to(torch.cuda.current_device())

        # Decoder
        self.decoder = HierarchicalDecoder(
            note_embedding_dim=self.note_embedding_dim,
            num_notes=self.num_notes,
            z_dim=self.latent_space_dim,
            num_layers=self.num_decoder_layers,
            rnn_hidden_size=self.decoder_hidden_size,
            dropout=self.decoder_dropout_prob,
            rnn_class=torch.nn.GRU,
        ).to(torch.cuda.current_device())

    def forward(self, measure_score_tensor: Variable, train=True):
        """
        Implements the forward pass of the VAE
        :param measure_score_tensor: torch Variable,
                (batch_size, measure_seq_length)
        :param measure_metadata_tensor: torch Variable,
                (batch_size, measure_seq_length, num_metadata)
        :param train: bool,
        :return: torch Variable,
                (batch_size, measure_seq_length, self.num_notes)
        """
        # check input
        seq_len = measure_score_tensor.size(1)
        assert seq_len == self.num_ticks_per_measure
        # compute output of encoding layer
        z_dist = self.encoder(measure_score_tensor)

        # sample from distribution
        z_tilde = z_dist.rsample()

        # compute prior
        prior_dist = distributions.Normal(
            loc=torch.zeros_like(z_dist.loc), scale=torch.ones_like(z_dist.scale)
        )
        z_prior = prior_dist.sample()

        # compute output of decoding layer
        weights, samples = self.decoder(
            z=z_tilde, score_tensor=measure_score_tensor, train=train
        )
        return weights, samples, z_dist, prior_dist, z_tilde, z_prior



import pytorch_lightning as pl
import torch
import torch.nn as nn
from torch.nn import functional as F
from torch.optim.lr_scheduler import CyclicLR


# Definir la red
class SketchNet_V2(nn.Module):
    def __init__(self, zp_dims, zr_dims, pf_num, inpaint_len, gen_dims, vae_model):
        super(SketchNet_V2, self).__init__()
        self.vae_model = vae_model

        self.zp_dims = zp_dims
        self.zr_dims = zr_dims
        self.pf_num = pf_num
        self.inpaint_len = inpaint_len
        self.gen_dims = gen_dims

        # Capa LSTM
        self.p_gru = nn.GRU(
            zp_dims * 2,
            gen_dims,
            pf_num,
            batch_first=True,
            bidirectional=True,
            dropout=0.5,
        )
        self.r_gru = nn.GRU(
            zr_dims * 2,
            gen_dims,
            pf_num,
            batch_first=True,
            bidirectional=True,
            dropout=0.5,
        )

        self.gen_p_gru = nn.GRU(
            zp_dims, gen_dims, pf_num, batch_first=True, bidirectional=True, dropout=0.5
        )
        self.gen_r_gru = nn.GRU(
            zr_dims, gen_dims, pf_num, batch_first=True, bidirectional=True, dropout=0.5
        )

        # Generation Output
        self.gen_p_out = nn.Linear(gen_dims * 2, zp_dims)
        self.gen_r_out = nn.Linear(gen_dims * 2, zr_dims)

        self.xavier_initialization()

    def gen_pitch_decoder(self, pf_pz, past_z, teacher_ratio, middle_z):
        y = past_z[:, -1, : self.zp_dims].unsqueeze(1)  # Get last measure in Z past
        hxx = pf_pz
        z_pitches = []

        for i in range(self.inpaint_len):
            y, hxx = self.gen_p_gru(y, hxx)
            y = y.contiguous().view(y.size(0), -1)
            y = self.gen_p_out(y)
            z_pitches.append(y)
            # Teacher forcing
            if self.training and torch.rand(1).item() < teacher_ratio:
                y = middle_z[:, i, : self.zp_dims]
            y = y.unsqueeze(1)

        return torch.stack(z_pitches, 1)

    def gen_rhythm_decoder(self, pf_z, past_z, teacher_ratio, middle_z):
        y = past_z[:, -1, self.zp_dims :].unsqueeze(1)  # Get last measure in Z
        hxx = pf_z
        z_rhythms = []
        for i in range(self.inpaint_len):
            y, hxx = self.gen_r_gru(y, hxx)
            y = y.contiguous().view(y.size(0), -1)
            y = self.gen_r_out(y)
            z_rhythms.append(y)
            # Teacher forcing
            if self.training and torch.rand(1).item() < teacher_ratio:
                y = middle_z[:, i, self.zp_dims :]
            y = y.unsqueeze(1)
        return torch.stack(z_rhythms, 1)

    def get_pf_pz(self, past_pz, future_pz):
        # Input: [64, 6, 128], [64, 6, 128] (batch, n_measures, Z_vector)
        # Output: [4, 64, 256?] -> Es la dimensión de la hidden del LSTM principal
        pz = torch.cat((past_pz, future_pz), 2)
        _, h = self.p_gru(pz)
        pf_pz = h
        return pf_pz

    # Concat Past and Future "Embedding" for rhythm
    def get_pf_rz(self, past_rz, future_rz):
        # Input: [64, 6, 128], [64, 6, 128] (batch, n_measures, Z_vector)
        # Output: [4, 64, 256?] -> Es la dimensión de la hidden del LSTM principal
        rz = torch.cat((past_rz, future_rz), 2)
        _, h = self.r_gru(rz)
        pf_rz = h
        return pf_rz

    def forward(self, past_x, future_x, middle_x):
        past_z = self.get_z_seq(past_x)
        future_z = self.get_z_seq(future_x)
        middle_z = self.get_z_seq(middle_x)

        past_pz = past_z[:, :, : self.zp_dims]
        past_rz = past_z[:, :, self.zp_dims :]
        future_pz = future_z[:, :, : self.zp_dims]
        future_rz = future_z[:, :, self.zp_dims :]

        pf_pz = self.get_pf_pz(past_pz, future_pz)
        pf_rz = self.get_pf_rz(past_rz, future_rz)

        # Teacher Forcing
        teacher_ratio = 0.5

        # Autoreggresive generation of new Z measures
        gen_pz = self.gen_pitch_decoder(pf_pz, past_z, teacher_ratio, middle_z)
        gen_rz = self.gen_rhythm_decoder(pf_rz, past_z, teacher_ratio, middle_z)
        gen_z = torch.cat((gen_pz, gen_rz), -1)
        gen_m = self.get_measure(gen_z)
        return gen_m

    def get_measure(self, z):
        dummy = torch.zeros((z.size(0), self.vae_model.seq_len)).long().cuda()
        ms = []
        for i in range(self.inpaint_len):
            m = self.vae_model.final_decoder(z[:, i, :], dummy, is_train=False)
            ms.append(m)
        return torch.stack(ms, 1)

    def get_z_seq(self, x):
        if x is None:
            return None
        # order: px, rx, len_x, nrx, gd
        px, _, len_x, nrx, gd = x
        batch_size = px.size(0)
        px = px.view(-1, self.vae_model.seq_len)
        nrx = nrx.view(-1, self.vae_model.seq_len, 3)
        len_x = len_x.view(-1)
        p_dis = self.vae_model.pitch_encoder(px, len_x)
        r_dis = self.vae_model.rhythm_encoder(nrx)
        zp = p_dis.rsample()
        zr = r_dis.rsample()
        z = torch.cat((zp, zr), -1)
        z = z.view(batch_size, -1, self.zr_dims + self.zp_dims)
        return z

    def inpaint(self, past, future):
        dummy = None
        logits = self(past, future, dummy)
        y_pred = logits.view(-1, logits.size(-1)).max(-1)[-1]
        return y_pred

    def xavier_initialization(self):
        layers = self._modules.keys()
        for layer in layers:
            if layer != "vae_model":
                for name, param in self._modules[layer].named_parameters():
                    if "weight" in name:
                        nn.init.xavier_normal_(param)


class LitSketchNet_V2(SketchNet_V2, pl.LightningModule):
    def __init__(
        self,
        zp_dims,
        zr_dims,
        pf_num,
        inpaint_len,
        gen_dims,
        vae_model,
        learn_rate: float,
        batch_size: int,
        sch_factor: float,
        sch_step: int,
        sch_mode: str,
        sch_gamma: float,
        sch_momentum: bool,
    ) -> None:
        super(SketchNet_V2, self).__init__()
        super(LitSketchNet_V2, self).__init__(
            zp_dims, zr_dims, pf_num, inpaint_len, gen_dims, vae_model
        )
        self.learn_rate = learn_rate
        self.batch_size = batch_size
        self.sch_factor = sch_factor
        self.sch_step = sch_step
        self.sch_mode = sch_mode
        self.sch_gamma = sch_gamma
        self.sch_momentum = sch_momentum
        self.save_hyperparameters()

        self.loss_fn = F.cross_entropy

    def training_step(self, batch, batch_idx):
        d, _ = batch
        label = d["inpaint_gd_whole"]
        recon_x = self(d["past_x"], d["future_x"], d["middle_x"])
        loss = self.loss_fn(
            recon_x.view(-1, recon_x.size(-1)), label.view(-1), reduction="mean"
        )
        self.log(
            "train_loss",
            loss,
            on_epoch=True,
            on_step=False,
            batch_size=self.batch_size,
        )
        return loss

    def validation_step(self, batch, batch_idx):
        d, _ = batch
        label = d["inpaint_gd_whole"]
        recon_x = self(d["past_x"], d["future_x"], d["middle_x"])
        return recon_x.view(-1, recon_x.size(-1)), label.view(-1)

    def validation_epoch_end(self, val_step_outputs):
        """
        Computes losses or metrics on validation dataloader
        when epoch ends.
        val_step_outputs: list of validation_step outputs during the epoch
        """
        model_out, labels = zip(*val_step_outputs)
        model_out = torch.cat(model_out)
        labels = torch.cat(labels)
        val_loss = self.loss_fn(model_out, labels)
        self.log("val_loss", val_loss)

        # Validation metrics
        # ...

        return val_loss

    def configure_optimizers(self):
        optimizer = torch.optim.AdamW(self.parameters(), lr=self.learn_rate)
        print(type(self.learn_rate), type(self.sch_factor))
        scheduler = CyclicLR(
            optimizer,
            base_lr=self.learn_rate,
            max_lr=self.learn_rate * self.sch_factor,
            step_size_up=self.sch_step,
            mode=self.sch_mode,
            gamma=self.sch_gamma,
            cycle_momentum=self.sch_momentum,
        )
        return (
            {
                "optimizer": optimizer,
                "lr_scheduler_config": {
                    "scheduler": scheduler,
                    "interval": "step",
                    "monitor": "val_loss",
                },
            },
        )



import json
import logging
import os
import random
from pathlib import Path

import numpy as np
import pandas as pd
from tqdm import tqdm

PROJECT_DIR = Path(__file__).resolve().parents[2]
DATA_DIR = os.path.join(PROJECT_DIR, "data")
RAW_DIR = os.path.join(DATA_DIR, "raw")
PROCESSED_DIR = os.path.join(DATA_DIR, "processed")
INTERIM_DIR = os.path.join(DATA_DIR, "interim")


class Note:
    def __init__(self, row, ticks_per_note):
        self.track = int(row["track"])
        self.measure = int(row["measure"])
        self.rel_pos = round(row["relative_position"] * ticks_per_note)
        self.duration = round(row["duration"] * ticks_per_note)
        self.pitch = int(row["pitch"])
        self.velocity = int(row["velocity"])
        self.tempo = int(row["tempo"])

    def to_noteseq(self):
        return np.array([self.pitch] + [128] * (self.duration - 1))

    def to_remi(self):
        return np.array(
            [
                self.measure,
                self.rel_pos,
                self.duration,
                self.pitch,
                self.velocity,
                self.tempo,
            ]
        )


def to_noteseq(df, ticks_per_note):
    overflow = []
    # groups = df.groupby(by=["track", "measure"]).groups.items()
    last_measures = df.groupby("track").max()["measure"].tolist()
    tracks_ids = list(df.groupby("track").groups.keys())
    tracks = [[] for x in tracks_ids]
    groups = [range(last_measures[track] + 1) for track in tracks_ids]
    for track_ix, measures in enumerate(groups):
        track_df = df[df["track"] == track_ix]
        for m in measures:
            seq = [129 for _ in range(ticks_per_note * 4)]
            notes_df = track_df[track_df["measure"] == m]
            if len(notes_df) <= 0:
                i = 0
                while len(overflow) > 0:
                    if i >= ticks_per_note * 4:
                        break
                    else:
                        seq[i] = overflow.pop(0)
                        i += 1
            else:
                for note in notes_df.iterrows():
                    n = Note(note[1], ticks_per_note)
                    noteseq = n.to_noteseq()
                    # if noteseq se paso pal compas del lado
                    i = 0
                    while len(overflow) > 0:
                        if i >= ticks_per_note * 4:
                            break
                        else:
                            seq[i] = overflow.pop(0)
                            i += 1
                    for i, ix in enumerate(range(n.rel_pos, n.rel_pos + n.duration)):
                        if ix >= ticks_per_note * 4:
                            overflow.append(noteseq[i])
                        else:
                            seq[ix] = noteseq[i]

            tracks[track_ix].append(np.array(seq))
    return np.array(tracks, dtype="object")


def to_remi(df, ticks_per_note):
    groups = df.groupby(by=["track", "measure"]).groups.items()
    tracks = [[] for x in range(list(groups)[-1][0][0] + 1)]
    for keys, indexs in groups:
        seq = []
        for note in df.loc[
            indexs,
        ].iterrows():
            n = Note(note[1], ticks_per_note)
            remi = n.to_remi()
            seq.append(remi)
        tracks[n.track].append(np.array(seq))
    return np.array(tracks, dtype="object")


def make_split(interim_files, dataset, n_context):
    files = os.listdir(interim_files)
    # Dos csvs: measures y contexts
    # filename, track, measure
    # filename, track, measure_st, measure_end

    measures = []
    contexts = []
    logging.info("Generating data splits...")

    for file in tqdm(files):
        npy_file = f"{file.split('.')[0]}.npy"
        path = os.path.join(interim_files, file)
        df = pd.read_csv(path)
        tracks = df["track"].unique()
        max_measures = [df[df["track"] == track]["measure"].max() for track in tracks]

        for track in tracks:
            for measure in range(max_measures[track]):

                measure_entry = {
                    "filename": npy_file,
                    "track": track,
                    "measure": measure,
                }
                measures.append(measure_entry)
                if measure + n_context <= max_measures[track]:
                    context_entry = {
                        "filename": npy_file,
                        "track": track,
                        "measure_st": measure,
                        "measure_end": measure + n_context,
                    }
                    contexts.append(context_entry)

    out_path_measures = os.path.join(DATA_DIR, "splits", f"{dataset}_measures.csv")
    out_path_contexts = os.path.join(DATA_DIR, "splits", f"{dataset}_contexts.csv")

    df_measures = pd.DataFrame(data=measures)
    df_contexts = pd.DataFrame(data=contexts)
    df_measures = add_train_val_test(df_measures)
    df_contexts = add_train_val_test(df_contexts)
    df_measures.to_csv(out_path_measures, index=False)
    df_contexts.to_csv(out_path_contexts, index=False)


def add_train_val_test(df):
    files = df["filename"].unique().tolist()
    random.Random(42).shuffle(files)
    n = len(files)
    train = [{"filename": f, "set": "train"} for f in files[: int(n * 0.8)]]
    val = [
        {"filename": f, "set": "val"}
        for f in files[len(train) : len(train) + int(n * 0.1)]
    ]
    test = [{"filename": f, "set": "test"} for f in files[len(train) + len(val) :]]

    train.extend(val)
    train.extend(test)

    files_sets = pd.DataFrame(data=train)
    merge = pd.merge(df, files_sets, on="filename")
    return merge


def main():
    logging.info("Creating necessary folders...")
    data_sources_file = open(os.path.join(RAW_DIR, "data_sources.json"), "r")
    sources = json.load(data_sources_file)
    for dataset, config in sources.items():
        interim_files = os.path.join(INTERIM_DIR, dataset)
        dataset_dir = os.path.join(PROCESSED_DIR, dataset)
        if not os.path.exists(dataset_dir):
            os.mkdir(dataset_dir)
        n_context = 16
        make_split(interim_files, dataset, n_context)
        for file in tqdm(os.listdir(interim_files)):
            path = os.path.join(interim_files, file)
            df = pd.read_csv(path)
            noteseq_16 = to_noteseq(df, 4)
            noteseq_24 = to_noteseq(df, 6)
            remi_16 = to_remi(df, 4)
            remi_24 = to_remi(df, 6)
            data = [noteseq_16, noteseq_24, remi_16, remi_24]
            formats = ["noteseq_16", "noteseq_24", "remi_16", "remi_24"]
            for f, d in zip(formats, data):
                out_dir = os.path.join(dataset_dir, f)
                if not os.path.exists(out_dir):
                    os.mkdir(out_dir)
                out_path = os.path.join(out_dir, f"{file.split('.')[0]}.npy")
                np.save(out_path, d, allow_pickle=True, fix_imports=False)


if __name__ == "__main__":
    log_fmt = "%(asctime)s - %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.INFO, format=log_fmt)
    main()



import hashlib
import json
import logging
import os
from pathlib import Path

import muspy
import pandas as pd
from tqdm import tqdm

PROJECT_DIR = Path(__file__).resolve().parents[2]
DATA_DIR = os.path.join(PROJECT_DIR, "data")
RAW_DIR = os.path.join(DATA_DIR, "raw")
INTERIM_DIR = os.path.join(DATA_DIR, "interim")


class StandardizedNote:
    def __init__(self, note, track, data, ticks_per_beat):
        self.track = track
        self.time = self.scaled_time(note, ticks_per_beat, data.resolution)
        self.relative_position = self.rel_pos(
            note.time, ticks_per_beat, data.barlines, data.resolution
        )
        self.duration = self.scaled_duration(note, ticks_per_beat, data.resolution)
        self.pitch = note.pitch
        self.velocity = note.velocity
        self.tempo = self.get_tempo(note, data.tempos)
        self.time_signature = self.get_time_signature(note, data.time_signatures)
        self.ticks_per_measure = self.get_ticks_per_measure(ticks_per_beat)
        self.measure = self.get_measure(note, data.barlines)

    def get_measure(self, note, barlines):
        measure = len(barlines) - len([m for m in barlines if note.time < m.time]) - 1
        return measure

    def scaled_time(self, note, ticks_per_beat, resolution):
        return round(note.time * ticks_per_beat / resolution) / ticks_per_beat

    def scaled_duration(self, note, ticks_per_beat, resolution):
        return round(note.duration * ticks_per_beat / resolution) / ticks_per_beat

    def get_tempo(self, note, tempos):
        last_tempo = sorted(
            [tempo for tempo in tempos if note.time >= tempo.time], key=lambda x: x.time
        )[-1]
        return last_tempo.qpm

    def get_time_signature(self, note, time_signatures):
        last_ts = sorted(
            [ts for ts in time_signatures if note.time >= ts.time], key=lambda x: x.time
        )[-1]
        return last_ts.numerator, last_ts.denominator

    def get_ticks_per_measure(self, ticks_per_beat):
        return int(self.time_signature[0] * 4 / self.time_signature[1])

    def rel_pos(self, time, ticks_per_beat, barlines, resolution):
        lower_time = [bar.time for bar in barlines if bar.time - time <= 0][-1]
        return round((time - lower_time) * ticks_per_beat / resolution) / ticks_per_beat


def get_hash(file_path):
    md5_hash = hashlib.md5()
    a_file = open(file_path, "rb")
    content = a_file.read()
    md5_hash.update(content)
    digest = md5_hash.hexdigest()
    return digest


def main():
    logging.info("Creating necessary folders...")
    data_sources_file = open(os.path.join(RAW_DIR, "data_sources.json"), "r")
    sources = json.load(data_sources_file)
    hashes = []
    for dataset, config in sources.items():
        raw_files = os.path.join(RAW_DIR, dataset)
        out_dir = os.path.join(INTERIM_DIR, dataset)
        if not os.path.exists(out_dir):
            os.mkdir(out_dir)
        for file in tqdm(os.listdir(raw_files)):
            path = os.path.join(raw_files, file)
            hash_ = get_hash(path)
            if hash_ not in hashes:
                hashes.append(hash_)
            else:
                continue
            try:
                if ".mxl" in file:
                    data = muspy.read_musicxml(path)
                if ".mid" in file:
                    data = muspy.read_midi(path)
            except IOError:
                continue

            if len(data.tempos) == 0:
                data.tempos = [muspy.Tempo(time=0.0, qpm=120)]
            # breakpoint()

            ticks_per_beat = 24
            data.infer_barlines()
            notes = [
                StandardizedNote(
                    note,
                    i,
                    data,
                    ticks_per_beat,
                )
                for i, track in enumerate(data.tracks)
                for note in track.notes
            ]

            notes = [
                [
                    note.track,
                    note.measure,
                    note.time,
                    note.relative_position,
                    note.duration,
                    note.pitch,
                    note.velocity,
                    note.tempo,
                    f"{note.time_signature[0]}/{note.time_signature[1]}",
                    note.ticks_per_measure,
                ]
                for note in notes
            ]

            df = pd.DataFrame(
                columns=[
                    # "resolution",
                    "track",
                    "measure",
                    "time",
                    "relative_position",
                    "duration",
                    "pitch",
                    "velocity",
                    "tempo",
                    "time_signature",
                    "ticks_per_measure",
                ],
                data=notes,
            )
            # Filters
            if len(df) == 0 or len(df["time_signature"].unique()) > 1:
                continue
            elif not (df["time_signature"].unique().item() == "4/4"):
                continue
            elif df["measure"].max().item() < 16:
                continue
            # Filtrar solo monofonias
            # elif len(df["track", "time"].unique()) != len(df):
            #    continue

            out_path = os.path.join(out_dir, f"{file.split('.')[0]}.csv")
            df.to_csv(out_path, index=False)


if __name__ == "__main__":
    log_fmt = "%(asctime)s - %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.INFO, format=log_fmt)
    main()



import os
from pathlib import Path

import pandas as pd
import plotly.express as px

PROJECT_DIR = Path(__file__).resolve().parents[2]
RESULTS_DIR = os.path.join(PROJECT_DIR, "results")

results_path = os.path.join(RESULTS_DIR, "metrics_results.csv")
df = pd.read_csv(results_path)


def get_plot(metric, name):
    fig = px.bar(
        df,
        x="Dataset",
        y=metric,
        color="Model",
        barmode="group",
        color_discrete_sequence=px.colors.qualitative.Bold,
        height=500,
        width=500,
        labels={metric: name, "Model": "Model", "Dataset": "Dataset"},
    )
    fig.update_layout(
        margin=dict(l=20, r=20, t=20, b=20), font=dict(size=18), showlegend=False
    )
    fig.write_image(os.path.join(RESULTS_DIR, f"images/{metric}.svg"))


metrics = ["pos_f1", "pAcc", "rAcc", "sdiv", "hdiv", "gsdiv"]
names = [
    "Position Score",
    "Pitch Accuracy",
    "Rhythm Accuracy",
    "Silence Divergence",
    "Pitch Class Divergence",
    "Groove Similarity Divergence",
]
[get_plot(m, n) for m, n in zip(metrics, names)]



import torch
from torch.nn.functional import one_hot

__all__ = ["Compose", "FixEmptyMeasures", "Factorize"]


class Compose:
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, midi):
        for t in self.transforms:
            midi = t(midi)
        return midi


class FixEmptyMeasures:
    def __init__(self, resolution):
        self.resolution = resolution

    def __call__(self, data):
        n = len(data) // self.resolution
        for i in range(n):
            measure = data[i * self.resolution : (i + 1) * self.resolution]
            if len(measure[measure < 128]) == 0:
                data[i * self.resolution] = 60
        return data


class Factorize:
    def __init__(self, ctxt_split=(6, 4, 6), resolution=24):
        self.ctxt_split = ctxt_split
        self.resolution = resolution

    def __call__(self, data):
        data = torch.Tensor(data).squeeze(0)
        if len(self.ctxt_split) == 1:
            re = self.factorize(data, self.resolution)
            re = [x.squeeze(0) for x in re]
        elif len(self.ctxt_split) == 3:
            ctxt_resolution = tuple([x * self.resolution for x in self.ctxt_split])
            past, middle, future = data.split(ctxt_resolution)
            past_x = self.factorize(past, self.resolution)
            middle_x = self.factorize(middle, self.resolution)
            future_x = self.factorize(future, self.resolution)
            re = {
                "inpaint_gd_whole": middle_x[4].contiguous().view(-1),  # middle_gd
                "past_x": past_x,
                "middle_x": middle_x,
                "future_x": future_x,
            }
        return re

    def factorize(self, data, resolution):
        ones = torch.ones_like(data) * 127
        rx = torch.where(data < 128, ones, data) - 126
        nrx = torch.stack(
            one_hot((rx - 1).to(torch.int64), num_classes=3).split(resolution)
        )
        rx = torch.stack(rx.split(resolution))
        gd = torch.stack(data.split(resolution))
        len_x = torch.Tensor([len(m[m < 128]) for m in gd])
        px = [m[m < 128] for m in gd]
        px = torch.stack(
            [torch.cat((x, torch.ones(resolution - len(x)) * 128)) for x in px]
        )
        return [
            px.to(torch.long),
            rx.to(torch.float),
            len_x.to(torch.long),
            nrx.to(torch.float),
            gd.to(torch.long),
        ]



# SketchVAE
import pytorch_lightning as pl
import torch
from torch import nn
from torch.distributions import Normal, kl_divergence
from torch.nn import Parameter
from torch.nn import functional as F
from torch.optim.lr_scheduler import CyclicLR


class SketchVAE_v2(nn.Module):
    def __init__(
        self,
        input_dims,
        p_input_dims,
        r_input_dims,
        hidden_dims,
        zp_dims,
        zr_dims,
        seq_len,
        beat_num,
        tick_num,
        decay,
    ):
        super(SketchVAE_v2, self).__init__()
        # pitch_encoder
        self.p_vocab_dims = 10
        self.p_layer_num = 2
        self.p_embedding = nn.Embedding(input_dims, self.p_vocab_dims)
        self.p_encoder_gru = nn.GRU(
            self.p_vocab_dims,
            hidden_dims,
            num_layers=self.p_layer_num,
            batch_first=True,
            bidirectional=True,
            dropout=0.2,
        )
        self.p_linear_mu = nn.Linear(hidden_dims * 2 * self.p_layer_num, zp_dims)
        self.p_linear_var = nn.Linear(hidden_dims * 2 * self.p_layer_num, zp_dims)
        # rhythm_encoder
        self.r_layer_num = 2
        self.r_encoder_gru = nn.GRU(
            r_input_dims,
            hidden_dims,
            num_layers=self.r_layer_num,
            batch_first=True,
            bidirectional=True,
            dropout=0.2,
        )
        self.r_linear_mu = nn.Linear(hidden_dims * 2 * self.r_layer_num, zr_dims)
        self.r_linear_var = nn.Linear(hidden_dims * 2 * self.r_layer_num, zr_dims)
        # hierarchical_decoder
        self.beat_layer_num = 2
        self.tick_layer_num = 2
        self.z_to_beat_hidden = nn.Sequential(
            nn.Linear(zr_dims + zp_dims, hidden_dims * self.beat_layer_num), nn.SELU()
        )
        self.beat_0 = Parameter(data=torch.zeros(1))
        self.beat_gru = nn.GRU(
            1,
            hidden_dims,
            num_layers=self.beat_layer_num,
            dropout=0.2,
            batch_first=True,
        )
        self.beat_to_tick_hidden = nn.Sequential(
            nn.Linear(hidden_dims, hidden_dims * self.tick_layer_num), nn.SELU()
        )
        self.beat_to_tick_input = nn.Sequential(
            nn.Linear(hidden_dims, hidden_dims), nn.SELU()
        )
        self.tick_0 = Parameter(data=torch.zeros(self.p_vocab_dims))
        self.d_embedding = nn.Embedding(input_dims, self.p_vocab_dims)
        self.tick_gru = nn.GRU(
            self.p_vocab_dims + hidden_dims,
            hidden_dims,
            num_layers=self.tick_layer_num,
            dropout=0.2,
            batch_first=True,
        )
        self.tick_to_note = nn.Sequential(nn.Linear(hidden_dims, input_dims), nn.ReLU())
        # parameter initialization
        self.input_dims = input_dims
        self.p_input_dims = p_input_dims
        self.r_input_dims = r_input_dims
        self.zr_dims = zr_dims
        self.zp_dims = zp_dims
        self.hidden_dims = hidden_dims
        self.seq_len = seq_len
        self.beat_num = beat_num
        self.tick_num = tick_num
        # teacher forcing hyperparameters
        self.iteration = 0
        self.eps = 1.0
        self.decay = torch.FloatTensor([decay])

    def pitch_encoder(self, px, len_x):
        px = self.p_embedding(px)
        s_len_x, s_len_idx = torch.sort(len_x, descending=True)
        _, re_len_idx = torch.sort(s_len_idx)
        s_px = px.index_select(0, s_len_idx)
        padding_px = torch.nn.utils.rnn.pack_padded_sequence(
            s_px, s_len_x.cpu(), batch_first=True
        )
        padding_px = self.p_encoder_gru(padding_px)[-1]
        padding_px = padding_px.transpose(0, 1).contiguous()
        padding_px = padding_px.view(padding_px.size(0), -1)
        n_px = padding_px.index_select(0, re_len_idx)
        p_mu = self.p_linear_mu(n_px)
        p_var = self.p_linear_var(n_px).exp_()
        p_dis = Normal(p_mu, p_var)
        return p_dis

    def rhythm_encoder(self, rx):
        rx = self.r_encoder_gru(rx)[-1]
        rx = rx.transpose(0, 1).contiguous()
        rx = rx.view(rx.size(0), -1)
        r_mu = self.r_linear_mu(rx)
        r_var = self.r_linear_var(rx).exp_()
        r_dis = Normal(r_mu, r_var)
        return r_dis

    def final_decoder(self, z, gd, is_train=True):
        gd = self.d_embedding(gd)
        beat_out = self.forward_beat(z)
        recon = self.forward_tick(beat_out, gd, is_train)
        return recon

    def forward_beat(self, z):
        batch_size = z.size(0)
        h_beat = self.z_to_beat_hidden(z)
        h_beat = h_beat.view(batch_size, self.beat_layer_num, -1)
        h_beat = h_beat.transpose(0, 1).contiguous()
        beat_input = self.beat_0.unsqueeze(0).expand(batch_size, self.beat_num, 1)
        beat_out, _ = self.beat_gru(beat_input, h_beat)
        return beat_out

    def forward_tick(self, beat_out, gd, is_train=True):
        ys = []
        batch_size = beat_out.size(0)
        tick_input = self.tick_0.unsqueeze(0).expand(batch_size, self.p_vocab_dims)
        tick_input = tick_input.unsqueeze(1)
        y = tick_input
        for i in range(self.beat_num):
            h_tick = self.beat_to_tick_hidden(beat_out[:, i, :])
            h_tick = h_tick.view(batch_size, self.tick_layer_num, -1)
            h_tick = h_tick.transpose(0, 1).contiguous()
            c_tick = self.beat_to_tick_input(beat_out[:, i, :]).unsqueeze(1)
            for j in range(self.tick_num):
                y = torch.cat((y, c_tick), -1)
                y, h_tick = self.tick_gru(y, h_tick)
                y = y.contiguous().view(y.size(0), -1)
                y = self.tick_to_note(y)
                ys.append(y)
                y = y.argmax(-1)
                y = self.d_embedding(y)
                if self.training and is_train:
                    p = torch.rand(1).item()
                    if p < self.eps:
                        y = gd[:, i * self.tick_num + j, :]
                    # update the eps after one batch
                    self.eps = self.decay / (
                        self.decay + torch.exp(self.iteration / self.decay)
                    )
                y = y.unsqueeze(1)
        return torch.stack(ys, 1)

    def forward(self, px, rx, len_x, gd):
        """
        px: [batch, seq_len, 1] with p_input number range
        rx: [batch, seq_len, r_input]
        len_x: [batch, 1] the efficient length of each pitch sequence
        gd: [batch, seq_len, 1] groundtruth of the melody sequence
        """
        if self.training:
            self.iteration += 1
        p_dis = self.pitch_encoder(px, len_x)
        r_dis = self.rhythm_encoder(rx)
        zp = p_dis.rsample()
        zr = r_dis.rsample()
        z = torch.cat((zp, zr), -1)
        recon = self.final_decoder(z, gd)
        return recon, p_dis, r_dis, self.iteration


def std_normal(shape):
    N = Normal(torch.zeros(shape), torch.ones(shape))
    if torch.cuda.is_available():
        N.loc = N.loc.cuda()
        N.scale = N.scale.cuda()
    return N


def vae_loss_function(recon, target, p_dis, r_dis, beta):
    CE = F.cross_entropy(recon.view(-1, recon.size(-1)), target, reduction="mean")
    normal1 = std_normal(p_dis.mean.size())
    normal2 = std_normal(r_dis.mean.size())
    KLD1 = kl_divergence(p_dis, normal1).mean()
    KLD2 = kl_divergence(r_dis, normal2).mean()
    max_indices = recon.view(-1, recon.size(-1)).max(-1)[-1]
    correct = max_indices == target
    acc = torch.sum(correct.float()) / target.size(0)
    return acc, CE + beta * (KLD1 + KLD2)


class LitSketchVAE_v2(SketchVAE_v2, pl.LightningModule):
    def __init__(
        self,
        input_dims,
        p_input_dims,
        r_input_dims,
        hidden_dims,
        zp_dims,
        zr_dims,
        seq_len,
        beat_num,
        tick_num,
        decay,
        batch_size,
        optimizer,
        scheduler,
    ):
        super(SketchVAE_v2, self).__init__()
        super(LitSketchVAE_v2, self).__init__(
            input_dims,
            p_input_dims,
            r_input_dims,
            hidden_dims,
            zp_dims,
            zr_dims,
            seq_len,
            beat_num,
            tick_num,
            decay,
        )
        self.batch_size = batch_size
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.save_hyperparameters()
        self.loss_fn = vae_loss_function

    def training_step(self, batch, batch_idx):
        px, rx, len_x, nrx, gd = batch
        target = gd

        recon, p_dis, r_dis, iteration = self(px, nrx, len_x, gd)
        _, loss = self.loss_fn(recon, target.view(-1), p_dis, r_dis, 0.1)
        self.log(
            "train_loss",
            loss,
            on_epoch=True,
            on_step=False,
            batch_size=self.batch_size,
        )
        return loss

    def validation_step(self, batch, batch_idx):
        px, rx, len_x, nrx, gd = batch
        out = self(px, nrx, len_x, gd)
        return out, gd

    def validation_epoch_end(self, val_step_outputs):
        model_out, target = zip(*val_step_outputs)
        final_val_loss = 0
        for i in range(len(model_out)):
            recon, p_dis, r_dis, iteration = model_out[i]
            target_ = target[i]
            _, val_loss = self.loss_fn(recon, target_.view(-1), p_dis, r_dis, 0.1)
            final_val_loss += val_loss
        final_val_loss = final_val_loss / len(model_out)
        self.log("val_loss", final_val_loss)

        # Aplicar metricas
        # Generar datos para hacer divergence??
        return val_loss

    def configure_optimizers(self):
        if self.optimizer.name == "AdamW":
            optimizer = torch.optim.AdamW(self.parameters(), lr=self.optimizer.lr)
        elif self.optimizer.name == "Adam":
            optimizer = torch.optim.Adam(self.parameters(), lr=self.optimizer.lr)

        if self.scheduler.name == "CyclicLR":
            scheduler = CyclicLR(
                optimizer,
                base_lr=self.scheduler.base_lr,
                max_lr=self.scheduler.max_lr,
                step_size_up=self.scheduler.step_size_up,
                step_size_down=self.scheduler.step_size_down,
                mode=self.scheduler.mode,
                gamma=self.scheduler.gamma,
                cycle_momentum=self.scheduler.cycle_momentum,
            )
        return (
            {
                "optimizer": optimizer,
                "lr_scheduler_config": {
                    "scheduler": scheduler,
                    "interval": "step",
                    "monitor": "val_loss",
                },
            },
        )



import torch
import muspy
import numpy as np
import pandas as pd
from torch.nn.functional import one_hot

class Note:
    def __init__(self, row, ticks_per_note):
        self.track = int(row["track"])
        self.measure = int(row["measure"])
        self.rel_pos = round(row["relative_position"] * ticks_per_note)
        self.duration = round(row["duration"] * ticks_per_note)
        self.pitch = int(row["pitch"])
        self.velocity = int(row["velocity"])
        self.tempo = int(row["tempo"])

    def to_noteseq(self):
        return np.array([self.pitch] + [128] * (self.duration - 1))

    def to_remi(self):
        return np.array(
            [
                self.measure,
                self.rel_pos,
                self.duration,
                self.pitch,
                self.velocity,
                self.tempo,
            ]
        )


class NoteSeq:
    def __init__(self, ticks_per_note):
        self.ticks_per_note = ticks_per_note
    
    def __call__(self, df):
        overflow = []
        # groups = df.groupby(by=["track", "measure"]).groups.items()
        last_measures = df.groupby("track").max()["measure"].tolist()
        tracks_ids = list(df.groupby("track").groups.keys())
        tracks = [[] for x in tracks_ids]
        groups = [range(last_measures[track] + 1) for track in tracks_ids]
        for track_ix, measures in enumerate(groups):
            track_df = df[df["track"] == track_ix]
            for m in measures:
                notes_df = track_df[track_df["measure"] == m]
                #time_signature = notes_df.iloc[0]["time_signature"].item()
                #num, den = time_signature.split("/")
                
                seq = [129 for _ in range(self.ticks_per_note * 4 * (num/den))]
                
                if len(notes_df) <= 0:
                    i = 0
                    while len(overflow) > 0:
                        if i >= self.ticks_per_note * 4:
                            break
                        else:
                            seq[i] = overflow.pop(0)
                            i += 1
                else:
                    for note in notes_df.iterrows():
                        n = Note(note[1], self.ticks_per_note)
                        noteseq = n.to_noteseq()
                        # if noteseq se paso pal compas del lado
                        i = 0
                        while len(overflow) > 0:
                            if i >= self.ticks_per_note * 4:
                                break
                            else:
                                seq[i] = overflow.pop(0)
                                i += 1
                        for i, ix in enumerate(range(n.rel_pos, n.rel_pos + n.duration)):
                            if ix >= self.ticks_per_note * 4:
                                overflow.append(noteseq[i])
                            else:
                                seq[ix] = noteseq[i]

                tracks[track_ix].append(np.array(seq))
        return np.array(tracks, dtype="object")



import torch
import muspy
import numpy as np
import pandas as pd
from torch.nn.functional import one_hot

__all__ = ["Compose", "FixEmptyMeasures", "Factorize", "Quantize"]


class Compose:
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, midi):
        for t in self.transforms:
            midi = t(midi)
        return midi


class FixEmptyMeasures:
    def __init__(self, resolution):
        self.resolution = resolution

    def __call__(self, data):
        # data shape: [num_measures, resolution]
        for measure in data:
            if len(measure[measure < 128]) == 0:
                measure[0] = 60
        return data


class Factorize:
    def __init__(self, ctxt_split=(6, 4, 6), resolution=24):
        self.ctxt_split = ctxt_split
        self.resolution = resolution

    def __call__(self, data):
        data = torch.Tensor(data).squeeze(0)
        if len(self.ctxt_split) == 1:
            re = self.factorize(data, self.resolution)
            re = [x.squeeze(0) for x in re]
        elif len(self.ctxt_split) == 3:
            past, middle, future = data.split(self.ctxt_split)
            past_x = self.factorize(past, self.resolution)
            middle_x = self.factorize(middle, self.resolution)
            future_x = self.factorize(future, self.resolution)
            re = {
                "inpaint_gd_whole": middle_x[4].contiguous().view(-1),  # middle_gd
                "past_x": past_x,
                "middle_x": middle_x,
                "future_x": future_x,
            }
        return re

    def factorize(self, data, resolution):
        data = data.reshape(-1)
        ones = torch.ones_like(data) * 127
        rx = torch.where(data < 128, ones, data) - 126
        nrx = torch.stack(
            one_hot((rx - 1).to(torch.int64), num_classes=3).split(resolution)
        )
        rx = torch.stack(rx.split(resolution))
        gd = torch.stack(data.split(resolution))
        len_x = torch.Tensor([len(m[m < 128]) for m in gd])
        px = [m[m < 128] for m in gd]
        px = torch.stack(
            [torch.cat((x, torch.ones(resolution - len(x)) * 128)) for x in px]
        )
        return [
            px.to(torch.long),
            rx.to(torch.float),
            len_x.to(torch.long),
            nrx.to(torch.float),
            gd.to(torch.long),
        ]

class StandardizedNote:
    def __init__(self, note, track, data, ticks_per_beat):
        self.track = track
        self.time = self.scaled_time(note, ticks_per_beat, data.resolution)
        self.relative_position = self.rel_pos(
            note.time, ticks_per_beat, data.barlines, data.resolution
        )
        self.duration = self.scaled_duration(note, ticks_per_beat, data.resolution)
        self.pitch = note.pitch
        self.velocity = note.velocity
        self.tempo = self.get_tempo(note, data.tempos)
        self.time_signature = self.get_time_signature(note, data.time_signatures)
        self.ticks_per_measure = self.get_ticks_per_measure(ticks_per_beat)
        self.measure = self.get_measure(note, data.barlines)

    def get_measure(self, note, barlines):
        measure = len(barlines) - len([m for m in barlines if note.time < m.time]) - 1
        return measure

    def scaled_time(self, note, ticks_per_beat, resolution):
        return round(note.time * ticks_per_beat / resolution) / ticks_per_beat

    def scaled_duration(self, note, ticks_per_beat, resolution):
        return round(note.duration * ticks_per_beat / resolution) / ticks_per_beat

    def get_tempo(self, note, tempos):
        last_tempo = sorted(
            [tempo for tempo in tempos if note.time >= tempo.time], key=lambda x: x.time
        )[-1]
        return last_tempo.qpm

    def get_time_signature(self, note, time_signatures):
        last_ts = sorted(
            [ts for ts in time_signatures if note.time >= ts.time], key=lambda x: x.time
        )[-1]
        return last_ts.numerator, last_ts.denominator

    def get_ticks_per_measure(self, ticks_per_beat):
        return int(self.time_signature[0] * 4 / self.time_signature[1])

    def rel_pos(self, time, ticks_per_beat, barlines, resolution):
        lower_time = [bar.time for bar in barlines if bar.time - time <= 0][-1]
        return round((time - lower_time) * ticks_per_beat / resolution) / ticks_per_beat



class Note:
    def __init__(self, row, ticks_per_note):
        self.track = int(row["track"])
        self.measure = int(row["measure"])
        self.rel_pos = round(row["relative_position"] * ticks_per_note)
        self.duration = round(row["duration"] * ticks_per_note)
        self.pitch = int(row["pitch"])
        self.velocity = int(row["velocity"])
        self.tempo = int(row["tempo"])

    def to_noteseq(self):
        return np.array([self.pitch] + [128] * (self.duration - 1))

    def to_remi(self):
        return np.array(
            [
                self.measure,
                self.rel_pos,
                self.duration,
                self.pitch,
                self.velocity,
                self.tempo,
            ]
        )

class Quantize:
    def __init__(self, ticks_per_beat):
        self.ticks_per_beat = ticks_per_beat


    def __call__(self, data):
        if len(data.tempos) == 0:
            data.tempos = [muspy.Tempo(time=0.0, qpm=120)]

        data.infer_barlines()
        notes = [
            StandardizedNote(
                note,
                i,
                data,
                self.ticks_per_beat,
            )
            for i, track in enumerate(data.tracks)
            for note in track.notes
        ]

        notes = [
            [
                note.track,
                note.measure,
                note.time,
                note.relative_position,
                note.duration,
                note.pitch,
                note.velocity,
                note.tempo,
                f"{note.time_signature[0]}/{note.time_signature[1]}",
                note.ticks_per_measure,
            ]
            for note in notes
        ]

        df = pd.DataFrame(
            columns=[
                # "resolution",
                "track",
                "measure",
                "time",
                "relative_position",
                "duration",
                "pitch",
                "velocity",
                "tempo",
                "time_signature",
                "ticks_per_measure",
            ],
            data=notes,
        )
        return df

class NoteSeq:
    def __init__(self, ticks_per_note):
        self.ticks_per_note = ticks_per_note
    
    def __call__(self, df):
        overflow = []
        # groups = df.groupby(by=["track", "measure"]).groups.items()
        last_measures = df.groupby("track").max()["measure"].tolist()
        tracks_ids = list(df.groupby("track").groups.keys())
        tracks = [[] for x in tracks_ids]
        groups = [range(last_measures[track] + 1) for track in tracks_ids]
        for track_ix, measures in enumerate(groups):
            track_df = df[df["track"] == track_ix]
            for m in measures:
                seq = [129 for _ in range(self.ticks_per_note * 4)]
                notes_df = track_df[track_df["measure"] == m]
                if len(notes_df) <= 0:
                    i = 0
                    while len(overflow) > 0:
                        if i >= self.ticks_per_note * 4:
                            break
                        else:
                            seq[i] = overflow.pop(0)
                            i += 1
                else:
                    for note in notes_df.iterrows():
                        n = Note(note[1], self.ticks_per_note)
                        noteseq = n.to_noteseq()
                        # if noteseq se paso pal compas del lado
                        i = 0
                        while len(overflow) > 0:
                            if i >= self.ticks_per_note * 4:
                                break
                            else:
                                seq[i] = overflow.pop(0)
                                i += 1
                        for i, ix in enumerate(range(n.rel_pos, n.rel_pos + n.duration)):
                            if ix >= self.ticks_per_note * 4:
                                overflow.append(noteseq[i])
                            else:
                                seq[ix] = noteseq[i]

                tracks[track_ix].append(np.array(seq))
        return np.array(tracks, dtype="object")



import os
from pathlib import Path

import numpy as np
import pandas as pd

from .transforms import Compose, Factorize, FixEmptyMeasures

PROJECT_DIR = Path(__file__).resolve().parents[2]
DATA_DIR = os.path.join(PROJECT_DIR, "data")
PROCESSED_DIR = os.path.join(DATA_DIR, "processed")
SPLIT_DIR = os.path.join(DATA_DIR, "splits")

transformations = {"Factorize": Factorize, "FixEmptyMeasures": FixEmptyMeasures}


class Dataset:
    def __init__(self, dataset, data_type, split, format_, resolution, transformations):
        path = os.path.join(SPLIT_DIR, f"{dataset}_{data_type}.csv")
        df = pd.read_csv(path)
        self.df = df[df["set"] == split].reset_index()
        self.dataset = dataset
        self.data_type = data_type
        self.format = format_
        self.resolution = resolution
        self.cache = {}
        self.transformations = Compose(
            [
                transformations[tf["name"]](*tf["params"], resolution=resolution)
                for tf in transformations
            ]
        )

    def __len__(self):
        return len(self.df)

    def __getitem__(self, idx):
        filename = self.df["filename"][idx]
        track = self.df["track"][idx]

        file_path = os.path.join(
            PROCESSED_DIR, self.dataset, f"{self.format}_{self.resolution}", filename
        )
        if filename not in self.cache:

            data = np.load(file_path, allow_pickle=True)
            if self.data_type == "measures":
                measure = self.df["measure"][idx]
                data = data[track][measure]
            elif self.data_type == "contexts":
                st_index = self.df["measure_st"][idx]
                end_index = self.df["measure_end"][idx]
                data = data[track][st_index:end_index]

            data = data.astype(int)
            data = self.transformations(data)
            self.cache[filename] = data
        else:
            data = self.cache[filename]
        return data



import json
import logging
import os
import shutil
from pathlib import Path

import gdown
import requests
from tqdm import tqdm

PROJECT_DIR = Path(__file__).resolve().parents[2]
DATA_DIR = os.path.join(PROJECT_DIR, "data")
EXTERNAL_DIR = os.path.join(DATA_DIR, "external")
RAW_DIR = os.path.join(DATA_DIR, "raw")
INTERIM_DIR = os.path.join(DATA_DIR, "interim")
PROCESSED_DIR = os.path.join(DATA_DIR, "processed")


def create_folders():
    if not os.path.exists(DATA_DIR):
        os.mkdir(DATA_DIR)
    if not os.path.exists(EXTERNAL_DIR):
        os.mkdir(EXTERNAL_DIR)
    if not os.path.exists(RAW_DIR):
        os.mkdir(RAW_DIR)
    if not os.path.exists(INTERIM_DIR):
        os.mkdir(INTERIM_DIR)
    if not os.path.exists(PROCESSED_DIR):
        os.mkdir(PROCESSED_DIR)


def download_progress_bar(url, out_file):
    r = requests.get(url, stream=True)
    total_size_in_bytes = int(r.headers.get("content-length", 0))
    block_size = 1024  # 1 Kibibyte
    progress_bar = tqdm(
        r.iter_content(), total=total_size_in_bytes, unit="iB", unit_scale=True
    )
    with open(out_file, "wb") as file:
        for data in r.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)
    progress_bar.close()
    if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
        print("ERROR, something went wrong")


def download_files(
    dataset_name,
    url,
    path,
    file_format,
    compression_format,
    gdownload=False,
    subdirectories=True,
):
    OUTPUT_FILE = os.path.join(RAW_DIR, f"{dataset_name}{compression_format}")
    if not gdownload:
        download_progress_bar(url, OUTPUT_FILE)
    else:
        gdown.download(url, OUTPUT_FILE, quiet=False)

    OUTPUT_DIR = os.path.join(RAW_DIR, dataset_name)
    TEMP_DIR = os.path.join(RAW_DIR, "temp")

    if not os.path.exists(TEMP_DIR):
        os.mkdir(TEMP_DIR)
    if not os.path.exists(OUTPUT_DIR):
        os.mkdir(OUTPUT_DIR)

    shutil.unpack_archive(OUTPUT_FILE, extract_dir=TEMP_DIR)
    for root, subdirs, files in tqdm(os.walk(TEMP_DIR)):
        [
            shutil.move(os.path.join(root, file), os.path.join(OUTPUT_DIR, file))
            for file in files
            if path in root and file_format in file
        ]
    shutil.rmtree(TEMP_DIR)
    os.remove(OUTPUT_FILE)


def main():
    logging.info("Creating necessary folders...")
    create_folders()
    data_sources_file = open(os.path.join(RAW_DIR, "data_sources.json"), "r")
    sources = json.load(data_sources_file)
    for dataset, config in sources.items():
        download_files(dataset, **config)


if __name__ == "__main__":
    log_fmt = "%(asctime)s - %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.INFO, format=log_fmt)
    main()
